@@ -1,9 +1,10 @@
 /*
- * Copyright 2001-2005 The Apache Software Foundation.
- *
- * Licensed under the Apache License, Version 2.0 (the "License")
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+ * Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -23,25 +24,31 @@ import java.util.Iterator;
 import java.util.List;
 import java.util.Set;
 
+import org.apache.commons.collections.CollectionUtils;
 import org.apache.commons.configuration.HierarchicalConfiguration.Node;
+import org.apache.commons.configuration.event.ConfigurationEvent;
+import org.apache.commons.configuration.event.ConfigurationListener;
+import org.apache.commons.configuration.tree.DefaultConfigurationNode;
+import org.apache.commons.configuration.tree.DefaultExpressionEngine;
+import org.apache.commons.configuration.tree.ExpressionEngine;
 
 import junit.framework.TestCase;
 
 /**
  * Test class for HierarchicalConfiguration.
- * 
- * @version $Id: TestHierarchicalConfiguration.java 234100 2005-08-20 18:18:16Z oheger $
+ *
+ * @version $Id: TestHierarchicalConfiguration.java 439648 2006-09-02 20:42:10Z oheger $
  */
 public class TestHierarchicalConfiguration extends TestCase
 {
     private static String[] tables = { "users", "documents" };
-    
+
     private static String[][] fields =
     {
         { "uid", "uname", "firstName", "lastName", "email" },
         { "docid", "name", "creationDate", "authorID", "version" }
     };
-        
+
     private HierarchicalConfiguration config;
 
     protected void setUp() throws Exception
@@ -62,7 +69,7 @@ public class TestHierarchicalConfiguration extends TestCase
         HierarchicalConfiguration.Node nodeTables = createNode("tables", null);
         for(int i = 0; i < tables.length; i++)
         {
-            HierarchicalConfiguration.Node nodeTable = createNode("table", null); 
+            HierarchicalConfiguration.Node nodeTable = createNode("table", null);
             nodeTables.addChild(nodeTable);
             HierarchicalConfiguration.Node nodeName = createNode("name", tables[i]);
             nodeTable.addChild(nodeName);
@@ -77,7 +84,7 @@ public class TestHierarchicalConfiguration extends TestCase
 
         config.getRoot().addChild(nodeTables);
     }
-    
+
     public void testSetRoot()
     {
         try
@@ -89,11 +96,34 @@ public class TestHierarchicalConfiguration extends TestCase
         {
             //ok
         }
-        
+
         config.setRoot(new HierarchicalConfiguration.Node("test"));
         assertTrue(config.isEmpty());
     }
 
+    public void testSetRootNode()
+    {
+        config.setRootNode(new DefaultConfigurationNode("testNode"));
+        assertNotSame("Same root node", config.getRootNode(), config.getRoot());
+        assertEquals("Wrong name of root node", "testNode", config.getRoot().getName());
+
+        config.setRootNode(new HierarchicalConfiguration.Node("test"));
+        assertSame("Wrong root node", config.getRootNode(), config.getRoot());
+    }
+
+    public void testSetRootNodeNull()
+    {
+        try
+        {
+            config.setRootNode(null);
+            fail("Could set null root node!");
+        }
+        catch(IllegalArgumentException iex)
+        {
+            //ok
+        }
+    }
+
     public void testIsEmpty()
     {
         assertFalse(config.isEmpty());
@@ -110,27 +140,27 @@ public class TestHierarchicalConfiguration extends TestCase
     {
         assertNull(config.getProperty("tables.table.resultset"));
         assertNull(config.getProperty("tables.table.fields.field"));
-        
+
         Object prop = config.getProperty("tables.table(0).fields.field.name");
         assertNotNull(prop);
         assertTrue(prop instanceof Collection);
         assertEquals(5, ((Collection) prop).size());
-        
+
         prop = config.getProperty("tables.table.fields.field.name");
         assertNotNull(prop);
         assertTrue(prop instanceof Collection);
         assertEquals(10, ((Collection) prop).size());
-        
+
         prop = config.getProperty("tables.table.fields.field(3).name");
         assertNotNull(prop);
         assertTrue(prop instanceof Collection);
         assertEquals(2, ((Collection) prop).size());
-        
+
         prop = config.getProperty("tables.table(1).fields.field(2).name");
         assertNotNull(prop);
         assertEquals("creationDate", prop.toString());
     }
-    
+
     public void testSetProperty()
     {
         config.setProperty("tables.table(0).name", "resources");
@@ -138,7 +168,7 @@ public class TestHierarchicalConfiguration extends TestCase
         config.setProperty("tables.table.name", "tab1,tab2");
         assertEquals("tab1", config.getString("tables.table(0).name"));
         assertEquals("tab2", config.getString("tables.table(1).name"));
-        
+
         config.setProperty("test.items.item", new int[] { 2, 4, 8, 16 });
         assertEquals(3, config.getMaxIndex("test.items.item"));
         assertEquals(8, config.getInt("test.items.item(2)"));
@@ -146,18 +176,18 @@ public class TestHierarchicalConfiguration extends TestCase
         assertEquals(6, config.getInt("test.items.item(2)"));
         config.setProperty("test.items.item(2)", new int[] { 7, 9, 11 });
         assertEquals(5, config.getMaxIndex("test.items.item"));
-        
+
         config.setProperty("test", Boolean.TRUE);
         config.setProperty("test.items", "01/01/05");
         assertEquals(5, config.getMaxIndex("test.items.item"));
         assertTrue(config.getBoolean("test"));
         assertEquals("01/01/05", config.getProperty("test.items"));
-        
+
         config.setProperty("test.items.item", new Integer(42));
         assertEquals(0, config.getMaxIndex("test.items.item"));
         assertEquals(42, config.getInt("test.items.item"));
     }
-    
+
     public void testClearProperty()
     {
         config.clearProperty("tables.table(0).fields.field(0).name");
@@ -168,14 +198,14 @@ public class TestHierarchicalConfiguration extends TestCase
         assertEquals("documents", config.getProperty("tables.table.name"));
         config.clearProperty("tables.table");
         assertEquals("documents", config.getProperty("tables.table.name"));
-        
+
         config.addProperty("test", "first");
         config.addProperty("test.level", "second");
         config.clearProperty("test");
         assertEquals("second", config.getString("test.level"));
         assertFalse(config.containsKey("test"));
     }
-    
+
     public void testClearTree()
     {
         Object prop = config.getProperty("tables.table(0).fields.field.name");
@@ -185,40 +215,40 @@ public class TestHierarchicalConfiguration extends TestCase
         assertNotNull(prop);
         assertTrue(prop instanceof Collection);
         assertEquals(4, ((Collection) prop).size());
-        
+
         config.clearTree("tables.table(0).fields");
         assertNull(config.getProperty("tables.table(0).fields.field.name"));
         prop = config.getProperty("tables.table.fields.field.name");
         assertNotNull(prop);
         assertTrue(prop instanceof Collection);
         assertEquals(5, ((Collection) prop).size());
-        
+
         config.clearTree("tables.table(1)");
         assertNull(config.getProperty("tables.table.fields.field.name"));
     }
-    
+
     public void testContainsKey()
     {
         assertTrue(config.containsKey("tables.table(0).name"));
         assertTrue(config.containsKey("tables.table(1).name"));
         assertFalse(config.containsKey("tables.table(2).name"));
-        
+
         assertTrue(config.containsKey("tables.table(0).fields.field.name"));
         assertFalse(config.containsKey("tables.table(0).fields.field"));
         config.clearTree("tables.table(0).fields");
         assertFalse(config.containsKey("tables.table(0).fields.field.name"));
-        
+
         assertTrue(config.containsKey("tables.table.fields.field.name"));
     }
-    
+
     public void testGetKeys()
     {
         List keys = new ArrayList();
         for (Iterator it = config.getKeys(); it.hasNext();)
         {
-            keys.add(it.next()); 
+            keys.add(it.next());
         }
-        
+
         assertEquals(2, keys.size());
         assertTrue(keys.contains("tables.table.name"));
         assertTrue(keys.contains("tables.table.fields.field.name"));
@@ -233,7 +263,7 @@ public class TestHierarchicalConfiguration extends TestCase
         assertEquals("2nd key", "order.key2", it.next());
         assertEquals("3rd key", "order.key3", it.next());
     }
-    
+
     public void testGetKeysString()
     {
         // add some more properties to make it more interesting
@@ -246,7 +276,7 @@ public class TestHierarchicalConfiguration extends TestCase
         config.addProperty("connections.connection.param.pwd", "secret");
         config.addProperty("connections.connection(-1).param.url", "url2");
         config.addProperty("connections.connection(1).param.user", "guest");
-        
+
         checkKeys("tables.table(1)", new String[] { "name", "fields.field.name" });
         checkKeys("tables.table(0)",
                 new String[] { "name", "fields.field.name", "tables.table(0)[@type]", "size", "fields.field.type", "fields.field.size" });
@@ -255,7 +285,7 @@ public class TestHierarchicalConfiguration extends TestCase
         checkKeys("connections.connection(1).param",
                 new String[] {"url", "user" });
     }
-    
+
     public void testAddProperty()
     {
         config.addProperty("tables.table(0).fields.field(-1).name", "phone");
@@ -263,7 +293,7 @@ public class TestHierarchicalConfiguration extends TestCase
         assertNotNull(prop);
         assertTrue(prop instanceof Collection);
         assertEquals(6, ((Collection) prop).size());
-        
+
         config.addProperty("tables.table(0).fields.field.name", "fax");
         prop = config.getProperty("tables.table.fields.field(5).name");
         assertNotNull(prop);
@@ -271,7 +301,7 @@ public class TestHierarchicalConfiguration extends TestCase
         List list = (List) prop;
         assertEquals("phone", list.get(0));
         assertEquals("fax", list.get(1));
-        
+
         config.addProperty("tables.table(-1).name", "config");
         prop = config.getProperty("tables.table.name");
         assertNotNull(prop);
@@ -286,17 +316,17 @@ public class TestHierarchicalConfiguration extends TestCase
         assertEquals(2, ((Collection) prop).size());
         assertEquals("confName",
         config.getProperty("tables.table(2).fields.field(1).name"));
-        
+
         config.addProperty("connection.user", "scott");
         config.addProperty("connection.passwd", "tiger");
         assertEquals("tiger", config.getProperty("connection.passwd"));
-        
+
         ConfigurationKey key = new ConfigurationKey();
         key.append("tables").append("table").appendIndex(0);
         key.appendAttribute("tableType");
         config.addProperty(key.toString(), "system");
         assertEquals("system", config.getProperty(key.toString()));
-        
+
         try
         {
             config.addProperty(".", "InvalidKey");
@@ -307,7 +337,7 @@ public class TestHierarchicalConfiguration extends TestCase
             //ok
         }
     }
-    
+
     public void testGetMaxIndex()
     {
         assertEquals(4, config.getMaxIndex("tables.table(0).fields.field"));
@@ -317,7 +347,7 @@ public class TestHierarchicalConfiguration extends TestCase
         assertEquals(0, config.getMaxIndex("tables.table(0).name"));
         assertEquals(0, config.getMaxIndex("tables.table(1).fields.field(1)"));
         assertEquals(-1, config.getMaxIndex("tables.table(2).fields"));
-        
+
         int maxIdx = config.getMaxIndex("tables.table(0).fields.field.name");
         for(int i = 0; i <= maxIdx; i++)
         {
@@ -326,7 +356,7 @@ public class TestHierarchicalConfiguration extends TestCase
             assertNotNull(config.getProperty(key.toString()));
         }
     }
-    
+
     public void testSubset()
     {
         // test the subset on the first table
@@ -337,7 +367,7 @@ public class TestHierarchicalConfiguration extends TestCase
         assertNotNull(prop);
         assertTrue(prop instanceof Collection);
         assertEquals(5, ((Collection) prop).size());
-        
+
         for (int i = 0; i < fields[0].length; i++)
         {
             ConfigurationKey key = new ConfigurationKey();
@@ -361,7 +391,91 @@ public class TestHierarchicalConfiguration extends TestCase
         subset = config.subset("tables.table.fields.field.name");
         assertTrue("subset is not empty", subset.isEmpty());
     }
-    
+
+    /**
+     * Tests the configurationAt() method to obtain a configuration for a sub
+     * tree.
+     */
+    public void testConfigurationAt()
+    {
+        HierarchicalConfiguration subConfig = config
+                .configurationAt("tables.table(1)");
+        assertEquals("Wrong table name", tables[1], subConfig.getString("name"));
+        List lstFlds = subConfig.getList("fields.field.name");
+        assertEquals("Wrong number of fields", fields[1].length, lstFlds.size());
+        for (int i = 0; i < fields[1].length; i++)
+        {
+            assertEquals("Wrong field at position " + i, fields[1][i], lstFlds
+                    .get(i));
+        }
+
+        subConfig.setProperty("name", "testTable");
+        assertEquals("Change not visible in parent", "testTable", config
+                .getString("tables.table(1).name"));
+        config.setProperty("tables.table(1).fields.field(2).name", "testField");
+        assertEquals("Change not visible in sub config", "testField", subConfig
+                .getString("fields.field(2).name"));
+    }
+
+    /**
+     * Tests the configurationAt() method when the passed in key does not exist.
+     */
+    public void testConfigurationAtUnknownSubTree()
+    {
+        try
+        {
+            config.configurationAt("non.existing.key");
+            fail("Could obtain sub config for unknown key!");
+        }
+        catch (IllegalArgumentException iex)
+        {
+            // ok
+        }
+    }
+
+    /**
+     * Tests the configurationAt() method when the passed in key selects
+     * multiple nodes. This should cause an exception.
+     */
+    public void testConfigurationAtMultipleNodes()
+    {
+        try
+        {
+            config.configurationAt("tables.table.name");
+            fail("Could create sub config with non unique key!");
+        }
+        catch (IllegalArgumentException iex)
+        {
+            // ok
+        }
+    }
+
+    /**
+     * Tests the configurationsAt() method.
+     */
+    public void testConfigurationsAt()
+    {
+        List lstFlds = config.configurationsAt("tables.table(1).fields.field");
+        assertEquals("Wrong size of fields", fields[1].length, lstFlds.size());
+        for (int i = 0; i < fields[1].length; i++)
+        {
+            HierarchicalConfiguration sub = (HierarchicalConfiguration) lstFlds
+                    .get(i);
+            assertEquals("Wrong field at position " + i, fields[1][i], sub
+                    .getString("name"));
+        }
+    }
+
+    /**
+     * Tests the configurationsAt() method when the passed in key does not
+     * select any sub nodes.
+     */
+    public void testConfigurationsAtEmpty()
+    {
+        assertTrue("List is not empty", config.configurationsAt("unknown.key")
+                .isEmpty());
+    }
+
     public void testClone()
     {
         Configuration copy = (Configuration) config.clone();
@@ -375,7 +489,26 @@ public class TestHierarchicalConfiguration extends TestCase
             }
         }
     }
-    
+
+    /**
+     * Tests whether registered event handlers are handled correctly when a
+     * configuration is cloned. They should not be registered at the clone.
+     */
+    public void testCloneWithEventListeners()
+    {
+        config.addConfigurationListener(new ConfigurationListener()
+        {
+            public void configurationChanged(ConfigurationEvent event)
+            {
+                // just a dummy
+            }
+        });
+        HierarchicalConfiguration copy = (HierarchicalConfiguration) config
+                .clone();
+        assertTrue("Event listener registered at clone", copy
+                .getConfigurationListeners().isEmpty());
+    }
+
     public void testAddNodes()
     {
         Collection nodes = new ArrayList();
@@ -387,18 +520,44 @@ public class TestHierarchicalConfiguration extends TestCase
         assertEquals("birthDate", config.getString("tables.table(0).fields.field(5).name"));
         assertEquals("lastLogin", config.getString("tables.table(0).fields.field(6).name"));
         assertEquals("language", config.getString("tables.table(0).fields.field(7).name"));
-        
+    }
+
+    /**
+     * Tests the addNodes() method when the provided key does not exist. In
+     * this case, a new node (or even a complete new branch) will be created.
+     */
+    public void testAddNodesForNonExistingKey()
+    {
+        Collection nodes = new ArrayList();
+        nodes.add(createNode("usr", "scott"));
+        Node nd = createNode("pwd", "tiger");
+        nd.setAttribute(true);
+        nodes.add(nd);
+        config.addNodes("database.connection.settings", nodes);
+
+        assertEquals("Usr node not found", "scott", config.getString("database.connection.settings.usr"));
+        assertEquals("Pwd node not found", "tiger", config.getString("database.connection.settings[@pwd]"));
+    }
+
+    /**
+     * Tests the addNodes() method when the new nodes should be added to an
+     * attribute node. This is not allowed.
+     */
+    public void testAddNodesWithAttributeKey()
+    {
+        Collection nodes = new ArrayList();
+        nodes.add(createNode("testNode", "yes"));
         try
         {
-            config.addNodes(".", nodes);
-            fail("Could use empty key!");
+            config.addNodes("database.connection[@settings]", nodes);
+            fail("Could add nodes to an attribute node!");
         }
         catch(IllegalArgumentException iex)
         {
             //ok
         }
     }
-    
+
     /**
      * Tests removing children from a configuration node.
      */
@@ -409,7 +568,7 @@ public class TestHierarchicalConfiguration extends TestCase
         assertFalse(node.hasChildren());
         node.removeChildren(); // should have no effect
         assertFalse(node.remove("child"));
-        
+
         node.addChild(createNode("test", "test"));
         assertTrue(node.hasChildren());
         assertTrue(node.remove("test"));
@@ -450,7 +609,52 @@ public class TestHierarchicalConfiguration extends TestCase
         assertEquals(28, v.beforeCount);
         assertEquals(v.beforeCount, v.afterCount);
     }
-    
+
+    /**
+     * Tests setting a custom expression engine, which uses a slightly different
+     * syntax.
+     */
+    public void testSetExpressionEngine()
+    {
+        config.setExpressionEngine(null);
+        assertNotNull("Expression engine is null", config.getExpressionEngine());
+        assertSame("Default engine is not used", HierarchicalConfiguration
+                .getDefaultExpressionEngine(), config.getExpressionEngine());
+
+        config.setExpressionEngine(createAlternativeExpressionEngine());
+        checkAlternativeSyntax();
+    }
+
+    /**
+     * Tests setting the default expression engine. This should impact all
+     * configuration instances that do not have their own engine.
+     */
+    public void testSetDefaultExpressionEngine()
+    {
+        ExpressionEngine engineOld = HierarchicalConfiguration.getDefaultExpressionEngine();
+        HierarchicalConfiguration
+                .setDefaultExpressionEngine(createAlternativeExpressionEngine());
+        checkAlternativeSyntax();
+        HierarchicalConfiguration.setDefaultExpressionEngine(engineOld);
+    }
+
+    /**
+     * Tests setting the default expression engine to null. This should not be
+     * allowed.
+     */
+    public void testSetDefaultExpressionEngineNull()
+    {
+        try
+        {
+            HierarchicalConfiguration.setDefaultExpressionEngine(null);
+            fail("Could set default expression engine to null!");
+        }
+        catch (IllegalArgumentException iex)
+        {
+            // ok
+        }
+    }
+
     /**
      * Helper method for testing the getKeys(String) method.
      * @param prefix the key to pass into getKeys()
@@ -463,7 +667,7 @@ public class TestHierarchicalConfiguration extends TestCase
         {
             values.add((expected[i].startsWith(prefix)) ? expected[i] :  prefix + "." + expected[i]);
         }
-        
+
         Iterator itKeys = config.getKeys(prefix);
         while(itKeys.hasNext())
         {
@@ -477,12 +681,57 @@ public class TestHierarchicalConfiguration extends TestCase
                 values.remove(key);
             }
         }
-        
+
         assertTrue("Remaining keys " + values, values.isEmpty());
     }
-    
+
+    /**
+     * Helper method for checking keys using an alternative syntax.
+     */
+    private void checkAlternativeSyntax()
+    {
+        assertNull(config.getProperty("tables/table/resultset"));
+        assertNull(config.getProperty("tables/table/fields/field"));
+
+        Object prop = config.getProperty("tables/table[0]/fields/field/name");
+        assertNotNull(prop);
+        assertTrue(prop instanceof Collection);
+        assertEquals(5, ((Collection) prop).size());
+
+        prop = config.getProperty("tables/table/fields/field/name");
+        assertNotNull(prop);
+        assertTrue(prop instanceof Collection);
+        assertEquals(10, ((Collection) prop).size());
+
+        prop = config.getProperty("tables/table/fields/field[3]/name");
+        assertNotNull(prop);
+        assertTrue(prop instanceof Collection);
+        assertEquals(2, ((Collection) prop).size());
+
+        prop = config.getProperty("tables/table[1]/fields/field[2]/name");
+        assertNotNull(prop);
+        assertEquals("creationDate", prop.toString());
+
+        Set keys = new HashSet();
+        CollectionUtils.addAll(keys, config.getKeys());
+        assertEquals("Wrong number of defined keys", 2, keys.size());
+        assertTrue("Key not found", keys.contains("tables/table/name"));
+        assertTrue("Key not found", keys
+                .contains("tables/table/fields/field/name"));
+    }
+
+    private ExpressionEngine createAlternativeExpressionEngine()
+    {
+        DefaultExpressionEngine engine = new DefaultExpressionEngine();
+        engine.setPropertyDelimiter("/");
+        engine.setIndexStart("[");
+        engine.setIndexEnd("]");
+        return engine;
+    }
+
     /**
      * Helper method for creating a field node with its children.
+     *
      * @param name the name of the field
      * @return the field node
      */
@@ -492,7 +741,7 @@ public class TestHierarchicalConfiguration extends TestCase
         fld.addChild(createNode("name", name));
         return fld;
     }
-    
+
     /**
      * Helper method for creating a configuration node.
      * @param name the node's name
@@ -505,7 +754,7 @@ public class TestHierarchicalConfiguration extends TestCase
         node.setValue(value);
         return node;
     }
-    
+
     /**
      * A test visitor implementation for checking whether all visitor methods
      * are correctly called.
